#!/anaconda3/bin/python
# -*- coding: utf-8 -*-

from AD.DualNumber import DualNumber
import AD.ElementaryFunctions as EF
import numpy as np
import pytest


def test_overload_add():
    x = DualNumber(5,1)
    y = DualNumber(7,1)

    assert (x + y).val == 12 and (x + y).der == 2
    assert (x + 5).val == 10 and (x + 5).der == 1
    assert (x + 5.0).val == 10 and (x + 5.0).der == 1
    assert (5 + x).val == 10 and (5 + x).der == 1
    assert (5.0 + x).val == 10 and (5.0 + x).der == 1


def test_overload_add_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x+"test"


def test_overload_multiply():
    x = DualNumber(5)
    y = DualNumber(7,2)

    assert (x * y).val == 35 and (x * y).der == 17
    assert (x * 5).val == 25 and (x * 5).der == 5
    assert (x * 5.0).val == 25 and (x * 5.0).der == 5
    assert (5.0 * x).val == 25 and (5.0 * x).der == 5
    assert (5 * x).val == 25 and (5 * x).der == 5


def test_overload_multiply_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x*"test"

def test_overload_sub():
    x = DualNumber(5,2)
    y = DualNumber(7,1)

    assert (x - y).val == -2 and (x - y).der == 1
    assert (x.__rsub__(y)).val == 2 and (x.__rsub__(y)).der == -1
    assert (x - 5).val == 0 and (x - 5).der == 2
    assert (x - 5.0).val == 0 and (x - 5.0).der == 2
    assert (5 - x).val == 0 and (5 - x).der == -2
    assert (5.0 - x).val == 0 and (5.0 - x).der == -2


def test_overload_sub_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x - "test"

def test_overload_truediv():
    x = DualNumber(10,2)
    y = DualNumber(2,4)
    print((x / y).der)
    assert (x / y).val == 5 and (x / y).der == (2*2-10*4)/(2**2)
    assert (x.__rtruediv__(y)).val == 0.2 and (x.__rtruediv__(y)).der == (4*10-2*2)/(10*10)
    assert (x / 5).val == 2 and (x / 5).der == 0.4
    assert (x / 5.0).val == 2.0 and (x / 5.0).der == 0.4
    assert (5 / x).val == 0.5 and (5 / x).der == -0.1
    assert (5.0 / x).val == 0.5 and (5.0 / x).der == -0.1


def test_overload_truediv_types():
    with pytest.raises(AssertionError):
      x = DualNumber(10)
      x/"test"

def test_overload_power():
    x = DualNumber(2,2)
    y = DualNumber(3,4)

    assert (x**y).val == 8 and (x**y).der == 8*(3/2*2+4*np.log(2))
    assert (x.__rpow__(y)).val == 9 and (x.__rpow__(y)).der == 9*(2/3*4+2*np.log(3))
    assert (x**2).val == 4 and (x**2).der == 8
    assert (x**2.0).val == 4 and (x**2.0).der == 8
    assert (2**x).val == 4 and (2**x).der == np.log(2)*4
    assert (2.0**x).val == 4 and (2.0**x).der == np.log(2)*4


def test_overload_power_types():
    with pytest.raises(AssertionError):
        x = DualNumber(5)
        x**"test"


def test_overload_pos():
    x = DualNumber(5,1)
    y = +x
    assert y.val == 5 and y.der == 1


def test_overload_neg():
    x = DualNumber(5,1)
    y = -x
    assert y.val == -5 and y.der == -1


def test_overload_abs():
    x = DualNumber(-5,-1)
    y = abs(x)
    assert y.val == 5 and y.der == 1
    x = DualNumber(5,1)
    y = abs(x)
    assert y.val == 5 and y.der == 1


def test_overload_round():
    x = DualNumber(5.123,1.0012)
    y = round(x,2)
    assert y.val == 5.12 and y.der == 1.00
    
def test_repr():
    x = DualNumber(5,1)
    assert repr(x) == 'Derivative: 1.00' + '\n' + 'Value: 5.00'
    
def test_str():
    x = DualNumber(5,1)
    assert print(x) == print('Derivative: 1.00' + '\n' + 'Value: 5.00')
    
def test_jacobian():
    x = DualNumber(5,1)
    assert x.jacobian == 1




def test_sin():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Sin(Test_Dual_Number_1).val == np.sin(1) and EF.Sin(Test_Dual_Number_1).der == np.cos(1)
    Test_Dual_Number_1 = 1
    assert EF.Sin(Test_Dual_Number_1).val == np.sin(1) and EF.Sin(Test_Dual_Number_1).der == 0

def test_cos():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Cos(Test_Dual_Number_1).val == np.cos(1) and EF.Cos(Test_Dual_Number_1).der == -np.sin(1)
    Test_Dual_Number_1 = 1
    assert EF.Cos(Test_Dual_Number_1).val == np.cos(1) and EF.Cos(Test_Dual_Number_1).der == 0

def test_tan():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Tan(Test_Dual_Number_1).val == np.tan(1) and EF.Tan(Test_Dual_Number_1).der == 1 / np.cos(1)**2
    Test_Dual_Number_1 = 1
    assert EF.Tan(Test_Dual_Number_1).val == np.tan(1) and EF.Tan(Test_Dual_Number_1).der == 0


def test_exp():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Exp(Test_Dual_Number_1).val == np.exp(1) and EF.Exp(Test_Dual_Number_1).der == np.exp(1)
    Test_Dual_Number_1 = 1
    assert EF.Exp(Test_Dual_Number_1).val == np.exp(1) and EF.Exp(Test_Dual_Number_1).der == 0

def test_Power():
    Test_Dual_Number_2 = DualNumber(2)
    assert EF.Power(Test_Dual_Number_2, 2).val == 4 and EF.Power(Test_Dual_Number_2, 2).der == 4
    Test_Dual_Number_2 = 2
    assert EF.Power(Test_Dual_Number_2, 2).val == 4 and EF.Power(Test_Dual_Number_2, 2).der == 0

def test_Log():
    Test_Dual_Number_1 = DualNumber(1)
    assert EF.Log(Test_Dual_Number_1).val == 0 and EF.Log(Test_Dual_Number_1).der == 1
    Test_Dual_Number_1 = 1
    assert EF.Log(Test_Dual_Number_1).val == 0 and EF.Log(Test_Dual_Number_1).der == 0

def test_ArcSin():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.ArcSin(Test_Dual_Number_1).val == np.arcsin(0.5) and EF.ArcSin(Test_Dual_Number_1).der == 1 / np.sqrt(
        0.75)
    Test_Dual_Number_1 = 0.5
    assert EF.ArcSin(Test_Dual_Number_1).val == np.arcsin(0.5) and EF.ArcSin(Test_Dual_Number_1).der == 0

def test_ArcCos():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.ArcCos(Test_Dual_Number_1).val == np.arccos(0.5) and EF.ArcCos(Test_Dual_Number_1).der == -1 / np.sqrt(
        0.75)
    Test_Dual_Number_1 = 0.5
    assert EF.ArcCos(Test_Dual_Number_1).val == np.arccos(0.5) and EF.ArcCos(Test_Dual_Number_1).der == 0

def test_ArcTan():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.ArcTan(Test_Dual_Number_1).val == np.arctan(0.5) and EF.ArcTan(Test_Dual_Number_1).der == 1 / 1.25
    Test_Dual_Number_1 = 0.5
    assert EF.ArcTan(Test_Dual_Number_1).val == np.arctan(0.5) and EF.ArcTan(Test_Dual_Number_1).der == 0

def test_Sqrt():
    Test_Dual_Number_1 = DualNumber(0.5)
    assert EF.Sqrt(Test_Dual_Number_1).val == np.sqrt(0.5) and EF.Sqrt(Test_Dual_Number_1).der == 1 / np.sqrt(
        0.5) / 2
    Test_Dual_Number_1 = 0.5
    assert EF.Sqrt(Test_Dual_Number_1).val == np.sqrt(0.5) and EF.Sqrt(Test_Dual_Number_1).der == 0

def test_data_type_check():
    x = 'x'
    try:
        EF.data_type_check(x)
    except AttributeError:
        return 0
    
    
    
    
test_overload_add()
test_overload_add_types()
test_overload_multiply()
test_overload_multiply_types()
test_overload_sub()
test_overload_sub_types()
test_overload_truediv()
test_overload_truediv_types()
test_overload_pos()
test_overload_neg()
test_overload_abs()
test_overload_round()
test_repr()
test_str()
test_jacobian()
test_sin()      
test_cos()
test_exp()
test_Power()
test_Log()
test_ArcSin()
test_ArcCos()
test_ArcTan()
test_Sqrt()
test_data_type_check()



#!/anaconda3/bin/python
# -*- coding: utf-8 -*-

from AD.DualNumber import DualNumber
import numpy as np

def Sin(x):
    '''
    >>> print(Sin(DualNumber(5,1)))
    Derivative: 0.28
    Value: -0.96
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.sin(x._val),np.cos(x._val)*x._der)
    else:
        return DualNumber(np.sin(x),0)

    
def Tan(x):
    '''
    >>> print(Tan(DualNumber(5,1)))
    Derivative: 12.43
    Value: -3.38
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.tan(x._val),(1+np.tan(x._val)*np.tan(x._val))*x._der)
    else:
        return DualNumber(np.tan(x),0)


def Cos(x):
    '''
    >>> print(Cos(DualNumber(5,1)))
    Derivative: 0.96
    Value: 0.28
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.cos(x._val),-1*np.sin(x._val)*x._der)
    else:
        return DualNumber(np.cos(x),0)

def Exp(x):
    '''
    >>> print(Exp(DualNumber(5,1)))
    Derivative: 148.41
    Value: 148.41
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.exp(x._val),np.exp(x._val)*x._der)
    else:
        return DualNumber(np.exp(x),0)

def Power(x,n):
    '''
    >>> print(Power(DualNumber(5,1),2))
    Derivative: 10.00
    Value: 25.00
    '''
    if data_type_check(x) == 0:
        return DualNumber(x._val**n,n*(x._val**(n-1))*x._der)
    else:
        return DualNumber(x**n,0)

def Log(x):
    '''
    >>> print(Log(DualNumber(5,1)))
    Derivative: 0.20
    Value: 1.61
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.log(x._val),(1/x._val)*x._der)
    else:
        return DualNumber(np.log(x),0)

        
def ArcSin(x):
    '''
    >>> print(ArcSin(DualNumber(0.5)))
    Derivative: 1.15
    Value: 0.52
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.arcsin(x._val),1/np.sqrt(1-x._val**2) * x._der)
    else:
        return DualNumber(np.arcsin(x),0)


def ArcCos(x):
    '''
    >>> print(ArcCos(DualNumber(0.5)))
    Derivative: -1.15
    Value: 1.05
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.arccos(x._val),-1/np.sqrt(1-x._val**2) * x._der)
    else:
        return DualNumber(np.arccos(x),0)
            
def ArcTan(x):
    '''
    >>> print(ArcTan(DualNumber(0.5)))
    Derivative: 0.80
    Value: 0.46
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.arctan(x._val),1/(1+x._val**2) * x._der)
    else:
        return DualNumber(np.arctan(x),0)

def Sqrt(x):
    '''
    >>> print(Sqrt(DualNumber(9,1)))
    Derivative: 0.17
    Value: 3.00
    '''
    if data_type_check(x) == 0:
        return DualNumber(np.sqrt(x._val),1/(2*np.sqrt(x._val)) * x._der)
    else:
        return DualNumber(np.sqrt(x),0)
        
def data_type_check(x):
   try:
       float(x._val)+float(x._der)
       return 0  # returns 0 if x is DualNumber
   except AttributeError:
       try:
           float(x)
           return 1 # returns 1 if x is real
       except:
           raise AttributeError('Input must be dual number or real number!')


if __name__ =="__main__":
    import doctest
    doctest.testmod()
        



#!/anaconda3/bin/python
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 13:20:32 2019

@author: shuvomsadhuka
"""

import numpy as np


class DualNumber():

    """Returns DualNumber and defines the overloading of arithmetric and unary operators.
    INPUTS
    =======
    real: float
      This stores the value of the function for which the derivative would be evaluated at.
    dual: float, optional, default value is 1
      This stores the value for the derivative of the function.
    RETURNS
    ========
    DualNumber: An instance with methods such as val() and der() to get out the current value and derivative of function respectively.
    NOTES
    =====
    PRE:
        - real, dual have numeric types
        - input of real is needed but input of real is optional
    POST:
        - real will be stored in the private attribute self._val and holds the current value of the function
        - dual will be stored in the private attribute self._der and holds the current derivative value of the function
        - raises a AttributeError exception if either real or dual is not of numeric type
        - returns a DualNumber instance
        - defines the overloading of arithmetic operators and unary operators
    EXAMPLES
    =========
    >>> [DualNumber(3,2).val, DualNumber(3,2).der]
    [3, 2]
    >>> [DualNumber(2.3,2).val, DualNumber(2.3,2).der]
    [2.3, 2]
    >>> [DualNumber(2,2.3).val, DualNumber(2,2.3).der]
    [2, 2.3]
    >>> [DualNumber(2.3,2.3).val, DualNumber(2.3,2.3).der]
    [2.3, 2.3]
    >>> print(DualNumber(2.3,2.3))
    Derivative: 2.30
    Value: 2.30
    """



    def __init__(self, real, dual=1):
        #print(type(real))
        assert (isinstance(real, int) or isinstance(real, float)), "Check the type of real!"
        assert (isinstance(dual, float) or isinstance(dual, int)), "Check the type of dual!"
        self._val = real
        self._der = dual

    @property
    def val(self):
        return self._val
    
    @property
    def der(self):
        return self._der
    
    @property
    def jacobian(self):
        return self._der
    
    def __repr__(self):
        return 'Derivative: {0:.2f}'.format(self.der) + '\n' + 'Value: {0:.2f}'.format(self.val)
        
    def __str__(self):
        return('Derivative: {0:.2f}'.format(self.der) + '\n' + 'Value: {0:.2f}'.format(self.val))

# Overloading arithmetic operators

    def __add__(self, other):
        '''
        >>> print(DualNumber(5,1)+3)
        Derivative: 1.00
        Value: 8.00
        '''
        try:
            val2 = self.val + other.val
            der2 = self.der + other.der
            return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = self.val + other
            der2 = self.der
            return DualNumber(val2, der2)

    def __radd__(self, other):
        '''
        >>> print(3+DualNumber(5,1))
        Derivative: 1.00
        Value: 8.00
        '''
        return self.__add__(other)

    def __mul__(self, other):
        '''
        >>> print(DualNumber(5,1)*3)
        Derivative: 3.00
        Value: 15.00
        '''
        try:
            val2 = self.val * other.val
            der2 = self.der * other.val + self.val * other.der
            return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = self.val * other
            der2 = self.der * other
            return DualNumber(val2, der2)

    def __rmul__(self, other):
        '''
        >>> print(3*DualNumber(5,1))
        Derivative: 3.00
        Value: 15.00
        '''
        return self.__mul__(other)

    def __sub__(self, other):
        '''
        >>> print(DualNumber(5,1)-3)
        Derivative: 1.00
        Value: 2.00
        '''
        try:
            val2 = self.val - other.val
            der2 = self.der - other.der
            return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = self.val - other
            der2 = self.der
            return DualNumber(val2, der2)

    def __rsub__(self, other):
        '''
        >>> print(3-DualNumber(5,1))
        Derivative: -1.00
        Value: -2.00
        '''
        try:
            val2 = other.val - self.val
            der2 = other.der - self.der
            return DualNumber(val2, der2)
        except AttributeError:
            assert(isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = other - self.val
            der2 = -self.der
            return DualNumber(val2, der2)

    def __truediv__(self, other):
        '''
        >>> print(DualNumber(5,1)/2)
        Derivative: 0.50
        Value: 2.50
        '''
        try:
            val2 = self.val / other.val
            der2 = (-self.val * other.der + self.der*other.val)/(other.val*other.val)
            return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = self.val / other
            der2 = self.der / other
            return DualNumber(val2, der2)

    def __rtruediv__(self, other):
        '''
        >>> print(2/DualNumber(5,1))
        Derivative: -0.08
        Value: 0.40
        '''
        try:
            val2 = other.val / self.val
            der2 = (other.der * self.val - other.val*self.der)/(self.val*self.val)
            return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = other / self.val
            der2 = -other*self.der / (self.val*self.val)
            return DualNumber(val2, der2)

    def __pow__(self, other):
        '''
        >>> print(DualNumber(5,1)**2)
        Derivative: 10.00
        Value: 25.00
        '''
        try:
            val2 = self.val ** other.val
            der2 = val2*(other.val/self.val*self.der+other.der*np.log(self.val))
            return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = self.val ** other
            der2 = other * (self.val ** (other - 1)) * self.der
            return DualNumber(val2, der2)


    def __rpow__(self, other):
        '''
        >>> print(2**DualNumber(5,1))
        Derivative: 22.18
        Value: 32.00
        '''
        try:
            val2 = other.val ** self.val
            der2 = val2*(self.val/other.val*other.der+self.der*np.log(other.val))
            return DualNumber(val2, der2)
        except AttributeError:
            assert (isinstance(other, float) or isinstance(other, int)), "Check the type of objects in function!"
            val2 = other ** self.val
            der2 = other ** self.val * np.log(other)
            return DualNumber(val2, der2)

# Overloading unary operators
    def __pos__(self):
        '''
        >>> print(+DualNumber(-5,-1))
        Derivative: -1.00
        Value: -5.00
        '''
        val2 = self.val
        der2 = self.der
        return DualNumber(val2, der2)

    def __neg__(self):
        '''
        >>> print(-DualNumber(-5,-1))
        Derivative: 1.00
        Value: 5.00
        '''
        val2 = -self.val
        der2 = -self.der
        return DualNumber(val2, der2)

    def __abs__(self):
        '''
        >>> print(abs(DualNumber(-5,-1)))
        Derivative: 1.00
        Value: 5.00
        '''
        val2 = abs(self.val)
        if self.val >= 0:
            der2 = self.der
        else:
            der2 = -1 * self.der
        return DualNumber(val2, der2)

    def __round__(self, n=None):
        '''
        >>> print(round(DualNumber(-5.234,-1.256),2))
        Derivative: -1.26
        Value: -5.23
        '''
        val2 = round(self.val, n)
        der2 = round(self.der, n)
        return DualNumber(val2, der2)

if __name__ =="__main__":
    # x=DualNumber(-2.578,-1.2345)
    # y=DualNumber(3,1)
    # print(x)
    # f=round(x,2)
    # print(f.val,f.der)
    import doctest
    doctest.testmod()
    
